#include <cassert>
#include <memory>
#include <vector>
#include <deque>

#include <iostream>

// данный класс выделяет новую память из кучи
// и помечает ненужные блоки памяти как свободные
// возврат памяти обратно ОС не происходит
class Arena
{
public:
    Arena() = default;
    ~Arena()
    {
        assert(m_allocations == 0);
    }

    void* allocate(std::size_t n)
    {
        if (n > m_available)
        {
            m_chunks.emplace_back(100500);
            m_available = m_chunks.back().size();
            m_memory = &m_chunks.back().front();
        }

        auto mem = m_memory;
        m_available -= n;
        m_memory += n;
        ++m_allocations;
        return mem;
    }

    // фрагметация при удалении не учитывается
    void deallocate(void* p, std::size_t n) {
        --m_allocations;
        auto mem = (unsigned char*)p;

        // освобожденная память будет доступна для использования,
        // только тогда, когда освобождаемый блок - граничит 
        // со свободной памятью.
        if (mem + n == m_memory)
        {
            m_memory = mem;
            m_available += n;
        }
    }

private:
    std::deque<std::vector<unsigned char>> m_chunks;
    std::size_t    m_available = 0;
    unsigned char* m_memory = nullptr;
    int            m_allocations = 0;
};

template <class T>
struct ArenaAllocator
{
    using value_type = T;

    using Traits = std::allocator_traits<ArenaAllocator<T>>;

    // конструкторы по умолчанию и копирования обязательно должны быть для совместимости с STL
    ArenaAllocator() = default;
    template<class U> ArenaAllocator(const ArenaAllocator<U>& other) : m_arena(other.m_arena) { }
    // а это уже вольная реализация
    explicit ArenaAllocator(Arena& arena): m_arena(&arena) { }
    ~ArenaAllocator() noexcept {}

    // libstdc++ требует следующие определения
    using size_type = typename std::allocator<T>::size_type;
    using difference_type = typename std::allocator<T>::difference_type;
    using pointer = typename std::allocator<T>::pointer;
    using const_pointer = typename std::allocator<T>::const_pointer;
    using reference = typename std::allocator<T>::reference;
    using const_reference = typename std::allocator<T>::const_reference;


    // описанные ниже ф-ии и внеклассовые методы сравнения - требуются для работы с аллокатором
    // стандартных контейнеров
    T* allocate(std::size_t n) { return (T*)m_arena->allocate(n * sizeof(T)); }
    void deallocate(T* p, std::size_t n) { m_arena->deallocate(p, n * sizeof(T)); }

    // вызовет размещающий new и передаст конструктору объекта столько параметров, сколько необходимо.
    template<class U, class... Args> void construct(U* p, Args&&... args) { std::allocator<T>().construct(p, std::forward<Args>(args)...); }
    // обёртка, которая вызывае p->~U()
    template<class U> void destroy(U* p) { std::allocator<T>().destroy(p); }

    // Современные реализациях STL требуеют наличие такой структуры
    // как обёртки аналогичной traits технике.
    template<class U> struct rebind { using other = ArenaAllocator<U>; };

    Arena* m_arena = nullptr;
};

template<class T, class U> bool operator==(const ArenaAllocator<T>& lhs, const ArenaAllocator<U>& rhs) { return lhs.m_arena == rhs.m_arena; }
template<class T, class U> bool operator!=(const ArenaAllocator<T>& lhs, const ArenaAllocator<U>& rhs) { return !(lhs == rhs); }

#include <deque>
#include <functional>
#include <list>
#include <map>
#include <memory>
#include <set>
#include <string>
#include <unordered_set>
#include <unordered_map>
#include <vector>


using a_string = std::basic_string<char, std::char_traits<char>, ArenaAllocator<char>>;

template <class T> using a_vector = std::vector<T, ArenaAllocator<T>>;
template <class T> using a_deque= std::deque<T, ArenaAllocator<T>>;
template <class T> using a_list = std::list<T, ArenaAllocator<T>>;
template <class K> using a_set = std::set<K, std::less<K>, ArenaAllocator<K>>;
template <class K, class V> using a_map = std::map<K, V, std::less<K>, ArenaAllocator<std::pair<const K, V>>>;
template <class K> using a_unordered_set = std::unordered_set<K, std::hash<K>, std::equal_to<K>, ArenaAllocator<K>>;
template <class K, class V> using a_unordered_map = std::unordered_map<K, std::hash<K>, std::equal_to<K>, ArenaAllocator<std::pair<const K, V>>>;


#include <cstring>
class X {
public:
    X():array(new char[1024]){}
    ~X(){ delete [] array; }
    // для формы вызова #1 обязательно наличие конструктора копирования, иначе, из-за штатного
    // произойдет double free corruption
    X(const X& rhs) { if (array) delete [] array; array = new char[1024]; memcpy(this->array, rhs.array, 1024); }
private:
    char * array;
};


int main()
{

    Arena arena;
    ArenaAllocator<char> arena_allocator(arena);

    a_string s_empty(arena_allocator);
    a_string s_123("123", arena_allocator);

    a_vector<int> v_int({1, 2, 3}, arena_allocator);

    // #1
    a_vector<X> v_x(42, X(), arena_allocator);
    // #2 для такой формы, конструктор копирования не обязателен, т.к. для объектов будет вызываться
    // умолчальный конструктор.
    //a_vector<X> v_x(42, arena_allocator);
    a_vector<a_string> v_str({s_empty, s_123}, arena_allocator);
    a_vector<a_string> v_str_copy(v_str, arena_allocator);
    a_deque<int> d_int({1, 2, 3}, arena_allocator);
    a_list<int> l_int({1, 2, 3}, arena_allocator);
    a_set<int> s_int({1, 2, 3}, std::less<int>{}, arena_allocator);
    a_map<a_string, int> m_str_int(arena_allocator);
    a_unordered_set<int> us_int(arena_allocator);

    auto p = std::allocate_shared<int>(arena_allocator, 123);

}
